import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.HashSet;

import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;

public class SortedSetMapDemo {
    
    public static void main(String[] args) {
        
        plainSetDemo();
        sortedSetDemo();
        sortedSetDemoComparator();
        sortedMapDemo();
    }

    public static void sortedMapDemo() {

        System.out.println("\nRunning sortedMapDemo...");
        // sorted map (tree map) stores the keys in the order defined by 
        // natural ordering of the keys (which is Person)
        SortedMap<Person, Double> salaries = new TreeMap<>();
        
        salaries.put(new Person("Jan Novak", "2014/11/20"), 10000.0);
        salaries.put(new Person("Pavel Novotny", "1970/11/20"), 35000.0);
        salaries.put(new Person("Ales Vacha", "1960/11/20"), 35000.0);
        salaries.put(new Person("Jan Novak", "1970/11/20"), 50000.0);
        salaries.put(new Person("Jan Novak", "2014/11/20"), 20000.0);

        double salary = salaries.get(new Person("Jan Novak", "2014/11/20"));
        System.out.println("Jan Novak salary="+salary);
        
        // the pair are iterated in the order of the keys
        for(Map.Entry<Person, Double> me: salaries.entrySet() ) {
            System.out.println("Entry "+me.getKey()+"=>"+me.getValue());
        }
        
    }

    public static void plainSetDemo() {
        
        System.out.println("\nRunning plainSetDemo...");
        
        // just a hashset, not sorted (tree) set
        Set<Person> people = new HashSet<>();
        
        // adding some people
        people.add(new Person("Jan Novak", "2014/11/20"));
        people.add(new Person("Jan Novak", "2014/11/20"));
        people.add(new Person("Jan Novak", "1970/11/20"));
        people.add(new Person("Jan Novak", "2014/10/21") );
        
        // is the person in the set?
        System.out.println("Contains?: "
           + people.contains(new Person("Jan Novak", "2014/10/21")));
    }
    
    public static void sortedSetDemo() {
        
        System.out.println("\nRunning sortedSetDemo (Comparable)...");
        
        // now a sorted set (implemented by tree set)
        SortedSet<Person> persons = new TreeSet<>();
        
        persons.add(new Person("Jan Novak", "2014/11/20"));
        persons.add(new Person("Pavel Novotny", "1970/11/20"));
        persons.add(new Person("Ales Vacha", "1960/11/20"));
        persons.add(new Person("Jan Novak", "1970/11/20"));

        // people are now iterated in a sorted order
        // defined by the natural ordering, i.e.
        // by the compareTo method in the class Person
        for(Person p: persons) {
            System.out.println("Person "+p);
        }
        

    }
        
    public static void sortedSetDemoComparator() {
        
        System.out.println("\nRunning sortedSetDemoComparator (DateOfBirthComparator)...");
        
        // people will be sorted in the order
        // defined by the DateOfBirthComparator
        // i.e. outside of the class Person
        SortedSet<Person> persons = new TreeSet<>(new DateOfBirthComparator());
        
        persons.add(new Person("Jan Novak", "2014/11/20"));
        persons.add(new Person("Pavel Novotny", "1970/11/20"));
        persons.add(new Person("Ales Vacha", "1960/11/20"));
        persons.add(new Person("Jan Novak", "1970/11/20"));

        for(Person p: persons) {
            System.out.println("Person "+p);
        }
        

    }
}
class DateOfBirthComparator implements Comparator<Person> {
    public int compare(Person a, Person b) {
        int comp = a.getDate().compareTo(b.getDate());
        if(comp == 0) {
            return a.getName().compareTo(b.getName());
        } else return comp;
    }
}
class DateOfBirthComparatorDescending extends DateOfBirthComparator {
    public int compare(Person a, Person b) {
        return -super.compare(a, b);
    }
}

class Person implements Comparable<Person> {
    private String name;
    private String date;
    public Person(String name, String date) { this.name = name; this.date = date; }
    public String toString() { return name+" born "+date; }
    public int compareTo(Person p) {
        int comp = name.compareTo(p.name);
        if(comp == 0) {
            return date.compareTo(p.date);
        } else return comp;
    }
    public boolean equals(Object obj) {
        if(obj instanceof Person) {
            Person p = (Person)obj;
            return name.equals(p.name) && date.equals(p.date);
        } else return false;
    }
    public int hashCode() {
        return name.hashCode() ^ date.hashCode();
    }
    public String getDate() { return date; }
    public String getName() { return name; }
}